<?php

namespace Drupal\eme\Commands;

use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\Extension;
use Drupal\Core\Extension\ModuleExtensionList;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Lock\LockBackendInterface;
use Drupal\Core\State\StateInterface;
use Drupal\Core\StreamWrapper\StreamWrapperManagerInterface;
use Drupal\eme\Eme;
use Drupal\eme\EmeProcessor;
use Drupal\eme\EmeProcessorException;
use Drush\Commands\DrushCommands;

/**
 * A Drush commandfile.
 *
 * In addition to this file, you need a drush.services.yml
 * in root of your module, and a composer.json file that provides the name
 * of the services file to use.
 *
 * See these files for an example of injecting Drupal services:
 *   - http://cgit.drupalcode.org/devel/tree/src/Commands/DevelCommands.php
 *   - http://cgit.drupalcode.org/devel/tree/drush.services.yml
 */
class EmeCommands extends DrushCommands {

  /**
   * The database lock object.
   *
   * @var \Drupal\Core\Lock\LockBackendInterface
   */
  protected $lock;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The bundle info service.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface
   */
  protected $bundleInfoService;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The state service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * The stream wrapper manager.
   *
   * @var \Drupal\Core\StreamWrapper\StreamWrapperManagerInterface
   */
  protected $streamWrapperManager;

  /**
   * The file system.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * Module extension list service.
   *
   * @var \Drupal\Core\Extension\ModuleExtensionList
   */
  protected $moduleExtensionList;

  /**
   * Info about discovered previous exports.
   *
   * @var array[]
   */
  protected $discoveredExports;

  /**
   * Construct a DrushBatch.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager.
   * @param \Drupal\Core\Lock\LockBackendInterface $lock
   *   The database lock object.
   * @param \Drupal\Core\State\StateInterface $state
   *   The state service.
   * @param \Drupal\Core\StreamWrapper\StreamWrapperManagerInterface $stream_wrapper_manager
   *   The stream wrapper manager.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system.
   * @param \Drupal\Core\Extension\ModuleExtensionList $module_list
   *   The module extension list service.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, EntityFieldManagerInterface $entity_field_manager, LockBackendInterface $lock, StateInterface $state, StreamWrapperManagerInterface $stream_wrapper_manager, FileSystemInterface $file_system, ModuleExtensionList $module_list) {
    parent::__construct();
    $this->entityTypeManager = $entity_type_manager;
    $this->entityFieldManager = $entity_field_manager;
    $this->lock = $lock;
    $this->state = $state;
    $this->streamWrapperManager = $stream_wrapper_manager;
    $this->fileSystem = $file_system;
    $this->moduleExtensionList = $module_list;
    $this->discoveredExports = array_reduce($this->moduleExtensionList->reset()->getList(), function (array $carry, Extension $extension) {
      if (
        property_exists($extension, 'info') &&
        is_array($extension->info) &&
        isset($extension->info['eme_settings'])
      ) {
        $module_name = $extension->getName();
        $path = preg_replace("/{$module_name}$/", '', $extension->subpath);
        $carry[$module_name] = $extension->info['eme_settings'] + [
          'human_name' => $extension->info['name'],
          'path' => trim($path, '/'),
        ];
      }
      return $carry;
    }, []);
  }

  /**
   * Exports content to migration.
   *
   * @param array $options
   *   An associative array of options whose values come from cli.
   *
   * @option types
   *   IDs of entity types to export, separated by commas.
   * @option from-preset
   *   Use the entity types from the last form submit (if available).
   * @option destination
   *   The destination of the module. Defaults to 'modules/custom'.
   * @option update
   *   The name of the module to update.
   *
   * @usage eme:export --types node,block_content --destination profiles/modules/custom
   *   Export all custom blocks, nodes and their dependencies to a new module at
   *   location DRUPAL_ROOT/profiles/modules/custom
   *
   * @usage eme:export --update eme_migrate
   *   Refresh the previously created export which module name is "eme_migrate".
   *
   * @command eme:export
   * @aliases emex
   */
  public function export(array $options = [
    'types' => NULL,
    'from-preset' => NULL,
    'destination' => 'modules/custom',
    'update' => NULL,
  ]) {
    $entities_to_export = NULL;
    $module_name = Eme::getModuleName();
    $human_name = Eme::getModuleHumanName();
    $migration_prefix = Eme::getMigrationPrefix();
    $migration_group = Eme::getMigrationGroup();
    $data_subdir = Eme::getDataSubdir();
    $file_subdir = Eme::getFileAssetsSubdir();
    $extract_to = !empty($options['destination'])
      ? trim($options['destination'], '/')
      : NULL;

    if ($options['types']) {
      $entities_to_export = array_filter(explode(',', $options['types']));
    }
    if ($options['from-preset']) {
      $entities_to_export = array_filter(\Drupal::state()->get(Eme::EXPORT_FORM_STATE, []));
    }

    if ($options['update']) {
      if (!array_key_exists($options['update'], $this->discoveredExports)) {
        $this->logger()->error(dt('The specified export module does not exist.'));
      }
      $module_name = $options['update'];
      $entities_to_export = $this->discoveredExports[$module_name]['types'];
      $human_name = $this->discoveredExports[$module_name]['human_name'] ?? $human_name;
      $migration_prefix = $this->discoveredExports[$module_name]['prefix'] ?? $migration_prefix;
      $migration_group = $this->discoveredExports[$module_name]['group'] ?? $migration_group;
      $data_subdir = $this->discoveredExports[$module_name]['data_subdir'] ?? $data_subdir;
      $file_subdir = $this->discoveredExports[$module_name]['file_subdir'] ?? $file_subdir;
      $extract_to = $this->discoveredExports[$module_name]['path'];
    }

    if (!is_array($entities_to_export)) {
      $this->logger()->error(dt('No entity types were provided.'));
      return;
    }

    if (empty($options['destination'])) {
      $this->logger()->error(dt('Destination must be provided.'));
    }

    $eme_processor = new EmeProcessor(
      $this->lock,
      $this->entityTypeManager,
      $this->entityFieldManager,
      $this->fileSystem,
      $this->moduleExtensionList,
      $entities_to_export,
      $module_name,
      $human_name,
      $migration_prefix,
      $migration_group,
      $data_subdir,
      $file_subdir,
      $extract_to
    );
    if ($eme_processor->alreadyProcessing()) {
      $this->logger()->error(dt('Another process may be exporting content already.'));
      return;
    }

    try {
      $batch = [
        'operations' => [],
        'title' => dt('Exporting content to migration'),
        'init_message' => dt('Starting content reference discovery.'),
        'progress_message' => dt('Completed step @current of @total.'),
        'error_message' => dt('Content export has encountered an error.'),
      ];
      foreach ($eme_processor->initialize() as $process_step) {
        $batch['operations'][] = [
          [get_class($this), 'processBatch'],
          [$eme_processor, $process_step],
        ];
      }
    }
    catch (EmeProcessorException $e) {
      // There are validation errors.
      $this->logger()->error(dt('Content cannot be exported: @exception-message', [
        '@exception-message' => $e->getMessage(),
      ]));
    }
    catch (\Exception $e) {
      // Unknown exception.
      $this->logger()->error(dt('Unknown exception: @exception-class with message: @exception-message', [
        '@exception-class' => get_class($e),
        '@exception-message' => $e->getMessage(),
      ]));
    }

    // 5. Add batch operations as new batch sets.
    batch_set($batch);
    // 6. Process the batch sets.
    drush_backend_batch_process();
    // 6. Show some information.
    $this->logger()->success(dt('Export finished.'));
  }

  /**
   * Processes the content export batch while persisting the processor.
   *
   * @param \Drupal\eme\EmeProcessor $eme_processor
   *   The content export processor.
   * @param string $process_step
   *   The process step (a method) to do.
   * @param array|\ArrayAccess $context
   *   The batch context.
   */
  public static function processBatch(EmeProcessor $eme_processor, $process_step, &$context): void {
    if (!isset($context['sandbox']['eme_processor'])) {
      $context['sandbox']['eme_processor'] = $eme_processor;
    }
    $active_processor = $context['sandbox']['eme_processor'];
    assert($active_processor instanceof EmeProcessor);
    $active_processor->doStep($process_step, $context);
  }

}
