<?php

namespace Drupal\eme\Form;

use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Extension\Extension;
use Drupal\Core\Extension\ModuleExtensionList;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\State\StateInterface;
use Drupal\Core\StreamWrapper\StreamWrapperManagerInterface;
use Drupal\Core\Url;
use Drupal\eme\Access\TemporarySchemeAccessCheck;
use Drupal\eme\Eme;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\ContentEntityType;
use Drupal\eme\EmeProcessor;
use Drupal\eme\EmeProcessorException;
use Drupal\Core\Lock\LockBackendInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;

/**
 * Form for starting a Content Entity to Migrations batch.
 */
class ExportForm extends FormBase {

  /**
   * The database lock object.
   *
   * @var \Drupal\Core\Lock\LockBackendInterface
   */
  protected $lock;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The bundle info service.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface
   */
  protected $bundleInfoService;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The state service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * The stream wrapper manager.
   *
   * @var \Drupal\Core\StreamWrapper\StreamWrapperManagerInterface
   */
  protected $streamWrapperManager;

  /**
   * The file system.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * Module extension list service.
   *
   * @var \Drupal\Core\Extension\ModuleExtensionList
   */
  protected $moduleExtensionList;

  /**
   * Info about discovered previous exports.
   *
   * @var array[]
   */
  protected $discoveredExports;

  /**
   * Construct the ExportForm.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Entity\EntityTypeBundleInfoInterface $entity_type_bundle_info
   *   The bundle info service.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager.
   * @param \Drupal\Core\Lock\LockBackendInterface $lock
   *   The database lock object.
   * @param \Drupal\Core\State\StateInterface $state
   *   The state service.
   * @param \Drupal\Core\StreamWrapper\StreamWrapperManagerInterface $stream_wrapper_manager
   *   The stream wrapper manager.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system.
   * @param \Drupal\Core\Extension\ModuleExtensionList $module_list
   *   The module extension list service.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, EntityTypeBundleInfoInterface $entity_type_bundle_info, EntityFieldManagerInterface $entity_field_manager, LockBackendInterface $lock, StateInterface $state, StreamWrapperManagerInterface $stream_wrapper_manager, FileSystemInterface $file_system, ModuleExtensionList $module_list) {
    $this->entityTypeManager = $entity_type_manager;
    $this->entityFieldManager = $entity_field_manager;
    $this->bundleInfoService = $entity_type_bundle_info;
    $this->lock = $lock;
    $this->state = $state;
    $this->streamWrapperManager = $stream_wrapper_manager;
    $this->fileSystem = $file_system;
    $this->moduleExtensionList = $module_list;
    $this->discoveredExports = array_reduce($this->moduleExtensionList->reset()->getList(), function (array $carry, Extension $extension) {
      if (
        property_exists($extension, 'info') &&
        is_array($extension->info) &&
        isset($extension->info['eme_settings'])
      ) {
        $module_name = $extension->getName();
        $path = preg_replace("/{$module_name}$/", '', $extension->subpath);
        $carry[$module_name] = $extension->info['eme_settings'] + [
          'human_name' => $extension->info['name'],
          'path' => trim($path, '/'),
        ];
      }
      return $carry;
    }, []);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('entity_type.bundle.info'),
      $container->get('entity_field.manager'),
      $container->get('lock.persistent'),
      $container->get('state'),
      $container->get('stream_wrapper_manager'),
      $container->get('file_system'),
      $container->get('extension.list.module')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'eme_export_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, $entity_type = NULL) {
    $temporary_stream_access = (new TemporarySchemeAccessCheck($this->streamWrapperManager))->access();
    if (!$temporary_stream_access->isAllowed()) {
      $form['info'] = [
        '#type' => 'item',
        '#markup' => $this->t('The temporary file directory isn\'t accessible. You must configure it for being able to export content. See the <a href=":system-file-settings-link">File system configuration form</a> for further info.', [
          ':system-file-settings-link' => Url::fromRoute('system.file_system_settings')->toString(),
        ]),
      ];
      return $form;
    }
    $types_to_exclude = Eme::getExcludedTypes();
    $content_entity_types = array_reduce($this->entityTypeManager->getDefinitions(), function (array $carry, EntityTypeInterface $definition) use ($types_to_exclude) {
      if (!$definition instanceof ContentEntityType) {
        return $carry;
      }
      $entity_type_id = $definition->id();
      if (in_array($entity_type_id, $types_to_exclude, TRUE)) {
        return $carry;
      }
      $carry[$definition->id()] = $this->t('@plural-label (<code>@provider</code>)', [
        '@plural-label' => ucfirst($definition->getPluralLabel()),
        '@provider' => $definition->getProvider(),
      ]);
      return $carry;
    }, []);
    ksort($content_entity_types);

    if (!empty($entity_type) && !empty($content_entity_types[$entity_type])) {
      $form['info'] = [
        '#type' => 'html_tag',
        '#tag' => 'h2',
        '#value' => $this->t('Export only @plural-label.', [
          '@plural-label' => $content_entity_types[$entity_type],
        ]),
      ];
    }
    elseif (!empty($this->discoveredExports)) {
      $update_options = [];
      foreach ($this->discoveredExports as $module_name => $eme_info) {
        $update_options[$module_name] = $this->t('@human-name (<code>@provider</code>); exported types: <code>@type-list</code>', [
          '@human-name' => $eme_info['human_name'],
          '@provider' => $module_name,
          '@type-list' => implode(', ', $eme_info['types']),
        ]);
      }
      $default_value = count($update_options) === 1 ? $module_name : 0;
      $form['existing_extensions'] = [
        '#title' => $this->t('Update previously exported extensions'),
        '#type' => 'radios',
        '#options' => $update_options + [0 => $this->t('Create new export')],
        '#empty_value' => 0,
        '#default_value' => $default_value,
      ];
    }

    $form['entities_to_export'] = [
      '#type' => 'checkboxes',
      '#options' => $content_entity_types,
      '#title' => $this->t('Select which content entities should be exported to a migration module'),
      '#default_value' => $this->state->get(Eme::EXPORT_FORM_STATE, []),
      '#states' => [
        'visible' => [
          ':input[name="existing_extensions"]' => ['value' => '0'],
        ],
        'enabled' => [
          ':input[name="existing_extensions"]' => ['value' => '0'],
        ],
      ],
    ];

    if (!empty($entity_type) && !empty($content_entity_types[$entity_type])) {
      $form['entities_to_export'] = [
        '#type' => 'hidden',
        '#value' => [$entity_type => $entity_type],
      ];
    }

    $form['actions'] = [
      '#type' => 'actions',
      'submit' => [
        '#type' => 'submit',
        '#op' => 'default',
        '#value' => $this->t('Start export'),
      ],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $exportables = $form_state->getValue('entities_to_export');
    $entities_to_export = array_values(
      array_filter($exportables)
    );
    $module_name = Eme::getModuleName();
    $human_name = Eme::getModuleHumanName();
    $migration_prefix = Eme::getMigrationPrefix();
    $migration_group = Eme::getMigrationGroup();
    $data_subdir = Eme::getDataSubdir();
    $file_subdir = Eme::getFileAssetsSubdir();
    $extract_to = NULL;

    $update = $form_state->getValue('existing_extensions');

    if ($update && array_key_exists($update, $this->discoveredExports) && array_key_exists('types', $this->discoveredExports[$update])) {
      $module_name = $update;
      $entities_to_export = $this->discoveredExports[$module_name]['types'];
      $human_name = $this->discoveredExports[$module_name]['human_name'] ?? $human_name;
      $migration_prefix = $this->discoveredExports[$module_name]['prefix'] ?? $migration_prefix;
      $migration_group = $this->discoveredExports[$module_name]['group'] ?? $migration_group;
      $data_subdir = $this->discoveredExports[$module_name]['data_subdir'] ?? $data_subdir;
      $file_subdir = $this->discoveredExports[$module_name]['file_subdir'] ?? $file_subdir;
      $extract_to = $this->discoveredExports[$module_name]['path'];
    }
    else {
      $this->state->set(Eme::EXPORT_FORM_STATE, $exportables);
    }

    $eme_processor = new EmeProcessor(
      $this->lock,
      $this->entityTypeManager,
      $this->entityFieldManager,
      $this->fileSystem,
      $this->moduleExtensionList,
      $entities_to_export,
      $module_name,
      $human_name,
      $migration_prefix,
      $migration_group,
      $data_subdir,
      $file_subdir,
      $extract_to
    );
    if ($eme_processor->alreadyProcessing()) {
      $this->messenger()->addError($this->t('Another process may be exporting content already.'));
      return;
    }

    try {
      $batch = [
        'operations' => [],
        'finished' => [get_class($this), 'finishBatch'],
        'title' => $this->t('Exporting content to migration'),
        'init_message' => $this->t('Starting content reference discovery.'),
        'progress_message' => $this->t('Completed step @current of @total.'),
        'error_message' => $this->t('Content export has encountered an error.'),
      ];
      foreach ($eme_processor->initialize() as $process_step) {
        $batch['operations'][] = [
          [get_class($this), 'processBatch'],
          [$eme_processor, $process_step],
        ];
      }

      batch_set($batch);
    }
    catch (EmeProcessorException $e) {
      // There are validation errors.
      $this->messenger()->addMessage($this->t('Content cannot be exported: @exception-message', [
        '@exception-message' => $e->getMessage(),
      ]), 'error');
    }
    catch (\Exception $e) {
      // Unknown exception.
      $this->messenger()->addMessage($this->t('Unknown exception: @exception-class with message: @exception-message', [
        '@exception-class' => get_class($e),
        '@exception-message' => $e->getMessage(),
      ]), 'error');
    }
  }

  /**
   * Processes the content export batch while persisting the processor.
   *
   * @param \Drupal\eme\EmeProcessor $eme_processor
   *   The content export processor.
   * @param string $process_step
   *   The process step to do.
   * @param array $context
   *   The batch context.
   */
  public static function processBatch(EmeProcessor $eme_processor, $process_step, array &$context): void {
    if (!isset($context['sandbox']['eme_processor'])) {
      $context['sandbox']['eme_processor'] = $eme_processor;
    }
    $active_processor = $context['sandbox']['eme_processor'];
    assert($active_processor instanceof EmeProcessor);
    $active_processor->doStep($process_step, $context);
  }

  /**
   * Finish batch.
   *
   * This function is a static function to avoid serializing the ConfigSync
   * object unnecessarily.
   */
  public static function finishBatch($success, $results, $operations) {
    if ($success) {
      \Drupal::messenger()->addMessage(t('Content export finished.'), 'status');
      $route_name = $results['redirect']
        ? 'eme.eme_export_download'
        : 'eme.eme_export_form';
      return new RedirectResponse(Url::fromRoute($route_name)->toString(), 307);
    }
    else {
      // An error occurred. "$operations" contains the operations which remain
      // "unprocessed".
      $error_operation = reset($operations);
      \Drupal::messenger()->addMessage(t('An error occurred while processing %error_operation with arguments: @arguments', [
        '%error_operation' => $error_operation[0],
        '@arguments' => print_r($error_operation[1], TRUE),
      ]), 'error');
    }
  }

}
